/*
 * Copyright (C) 2010 Bjoern Biesenbach <bjoern@bjoern-b.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>
#include <stdio.h>


#define DS1371_ADDRESS 0xD0

// bits of control register
#define NEOSC 7
#define WACE 6
#define WDNALM 5
#define INTCN 3
#define RS2 2
#define RS1 1
#define AIE 0

// bits of status register
#define AF 0
#define OSF 7

static uint8_t twi_read_byte_ack(void){
    
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
    while(!(TWCR & (1<<TWINT))); 
    return TWDR;
}

static uint8_t twi_read_byte_nack(void){
    
    TWCR = (1<<TWINT)|(1<<TWEN);
    while(!(TWCR & (1<<TWINT))); 
    return TWDR;
}

static void twi_start(void){

    TWBR = 2; // 400kHz
    TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    while(!(TWCR & (1<<TWINT)));
}

static void twi_stop(void){

    TWCR = (1<<TWSTO)|(1<<TWINT)|(1<<TWEN);
    while(TWCR & (1<<TWINT));
    TWCR = 0;
}

static void twi_write_byte(uint8_t byte){
    
    TWDR = byte;
    TWCR = (1<<TWINT)|(1<<TWEN);
    while(!(TWCR & (1<<TWINT)));
}

static void twi_rw_operation(uint8_t sla, uint8_t pointer){

    twi_start();
    twi_write_byte(sla | TW_WRITE);
    twi_write_byte(pointer);
    twi_start();
    twi_write_byte(sla | TW_READ);
}

void ds1371_set_rtc(uint32_t time){

    twi_start();
    twi_write_byte(DS1371_ADDRESS | TW_WRITE);
    twi_write_byte(0); // TOD counter starts at address 0x00
    twi_write_byte((uint8_t)(time & 0xFF));
    twi_write_byte((uint8_t)((time>>8) & 0xFF));
    twi_write_byte((uint8_t)((time>>16) & 0xFF));
    twi_write_byte((uint8_t)((time>>24) & 0xFF));
    twi_stop();
}

void ds1371_set_alarm_time(uint32_t time){

    twi_start();
    twi_write_byte(DS1371_ADDRESS | TW_WRITE);
    twi_write_byte(4); // Alarm counter starts at address 0x04
    twi_write_byte((uint8_t)(time & 0xFF));
    twi_write_byte((uint8_t)((time>>8) & 0xFF));
    twi_write_byte((uint8_t)((time>>16) & 0xFF));
    twi_stop();
}

void ds1371_start_alarm_timer(void){

    twi_start();
    twi_write_byte(DS1371_ADDRESS | TW_WRITE);
    twi_write_byte(7); // address of control register
    twi_write_byte((1<<RS1)|(1<<RS2)|(1<<WACE)|(1<<INTCN)|(1<<AIE));
    twi_stop();
}

void ds1371_clear_alarm_timer(void){

    twi_start();
    twi_write_byte(DS1371_ADDRESS | TW_WRITE);
    twi_write_byte(8); // address of status register
    twi_write_byte(0);
    twi_stop();
}

uint32_t ds1371_read_rtc(void){
    
    uint8_t bytes[4];
    uint32_t time_of_day;
    char value[64];

    twi_rw_operation(DS1371_ADDRESS, 0);
    bytes[0] = twi_read_byte_ack();
    bytes[1] = twi_read_byte_ack();
    bytes[2] = twi_read_byte_ack();
    bytes[3] = twi_read_byte_nack();
    twi_stop();
    
    time_of_day = (uint32_t)bytes[0] | ((uint32_t)bytes[1]<<8) | ((uint32_t)bytes[2]<<16) | ((uint32_t)bytes[3]<<24);

    sprintf(value,"%ld",time_of_day);
    
    return 0;
}

